package com.mathieubolla.externalui;

import java.io.IOException;
import java.io.OutputStream;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingQueue;

import org.apache.commons.io.IOUtils;
import org.simpleframework.http.Request;
import org.simpleframework.http.Response;
import org.simpleframework.http.core.Container;
import org.simpleframework.transport.connect.Connection;
import org.simpleframework.transport.connect.SocketConnection;

import com.mathieubolla.externalui.events.Event;
import com.mathieubolla.externalui.events.EventListener;
import com.mathieubolla.externalui.events.EventProvider;
import com.mathieubolla.externalui.utils.StreamProvider;

public class ExternalUiProvider implements Container, EventProvider, EventListener {
	private Set<EventListener> listeners = new HashSet<EventListener>();
	private final StreamProvider provider;
	private BlockingQueue<Event> events = new LinkedBlockingQueue<Event>();
	private ExecutorService EXECUTORS = Executors.newCachedThreadPool();
	
	@Override
	public void handle(Request request, Response response) {
		OutputStream stream = null;
		try {
			stream = response.getOutputStream();
			
			if (request.getPath().getPath().equals("/update")) {
				sendUpdates(stream);
			} else {
				if (request.getPath().getPath().equals("/event")){
					decodeEvent(request, stream);
				} else {
					sendUi(stream);
				}
			}
		} catch (Throwable e) {
			e.printStackTrace();
		}
	}

	private void sendUi(OutputStream stream) throws IOException {
		IOUtils.copy(provider.get(), stream);
		IOUtils.closeQuietly(stream);
	}

	private void decodeEvent(Request request, OutputStream stream) throws IOException {
		String eventCode = request.getParameter("event");
		if (eventCode != null) {
			String[] split = eventCode.split(":");
			Event event = new Event(split[0], split[1]);
			for (EventListener listener : listeners) {
				listener.notify(event);
			}
		}
		IOUtils.closeQuietly(stream);
	}

	private Future<?> sendUpdates(OutputStream stream) {
		return EXECUTORS.submit(asyncSendUpdates(stream));
	}

	private Runnable asyncSendUpdates(final OutputStream stream) {
		return new Runnable() {
			@Override
			public void run() {
				try {
					if (events.size() > 1) {
						List<Event> drainedEvents = new ArrayList<Event>();
						events.drainTo(drainedEvents);
						for (Event event : drainedEvents) {
							IOUtils.write(event.getCode()+"="+event.getValue()+"&", stream);
						}
					} else {
						Event event = events.take();
						IOUtils.write(event.getCode()+"="+event.getValue(), stream);
					}
					IOUtils.closeQuietly(stream);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		};
	}
	
	@Override
	public void register(EventListener listener) {
		listeners.add(listener);
	}
	
	@Override
	public void unregister(EventListener listener) {
		listeners.remove(listener);
	}
	
	@Override
	public void notify(Event event) {
		System.out.println("Notified of "+event);
		System.out.println("Current event stack is "+events.size()+" long.");
		try {
			events.put(event);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}
	
	public ExternalUiProvider(int port, StreamProvider provider) {
		this.provider = provider;
		try {
			init(this, port);
		} catch (IOException e) {
			throw new RuntimeException("Can't startup external ui", e);
		}
	}
	
	private static void init(ExternalUiProvider container, int port) throws IOException {
		Connection connection = new SocketConnection(container);
		SocketAddress address = new InetSocketAddress(8080);
		
		connection.connect(address);
	}
}
